/*
 * Copyright (c) 2009,2010 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <pthread.h>

#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>


#include <linux/joystick.h>

#include <portLib.h>
#include <taskLib.h>

#include "rmp/rmpLib.h"

/* For joystick driver */
#define NAME_LENGTH 128

unsigned char axes = 2;		/* number of joystick axes */
unsigned char buttons = 2;	/* number of joystick buttons */
int *axis;			/* array of joystick axis values */
int *button;			/* array of joystick button state */
short v, w;			/* linear & angular speed references */
int quit;			/* quit flag */

int version = 0x000800;
char name[NAME_LENGTH] = "Unknown";

pthread_mutex_t joymutex = PTHREAD_MUTEX_INITIALIZER;

/* Speed gains for joystick control */
#define VGAIN (0.008)	/* linear  */
#define WGAIN (-0.016)	/* angular */


/*----------------------------------------------------------------------*/
/**
 ** Init joystick
 **/
static int
joyinit(char *dev)
{
	int joy;

	joy = open(dev, O_RDONLY);
	if (joy < 0) {
		perror(dev);
		return -1;
	}
	ioctl(joy, JSIOCGVERSION, &version);
	ioctl(joy, JSIOCGAXES, &axes);
	ioctl(joy, JSIOCGBUTTONS, &buttons);
	ioctl(joy, JSIOCGNAME(NAME_LENGTH), name);

	printf("Joystick (%s) has %d axes and %d buttons.\n"
	       "\tDriver version is %d.%d.%d.\n",
	       name, axes, buttons, version >> 16,
	       (version >> 8) & 0xff, version & 0xff);
	return joy;
}
/*----------------------------------------------------------------------*/

/**
 ** task reading data from the joystick and producing references;
 **/
int
joyRead(int fd)
{
	struct js_event js;
	struct pollfd pfd;
	int i, ret, nread;

	while (!quit) {
		pfd.fd = fd;
		pfd.events = POLLIN;
		pfd.revents = 0;
		ret = poll(&pfd, 1, 100);

		if (ret < 0) {
			if (errno != EINTR)
				fprintf(stderr, "poll error: %s\n",
					strerror(errno));
			continue;
		}
		if (ret == 0)
			/* nothing to read from joystick */
			continue;
		if (pfd.revents & POLLIN != POLLIN)
			continue;
		/* something to read from joystick */
		if (read(fd, &js, sizeof(struct js_event))
		    != sizeof(struct js_event)) {
			perror("\njstest: error reading");
			break;
		}
		switch(js.type & ~JS_EVENT_INIT) {
		case JS_EVENT_BUTTON:
			button[js.number] = js.value;
			printf("button %d %d\n", js.number, js.value);
			break;
		case JS_EVENT_AXIS:
			axis[js.number] = js.value;
			printf("axis %d %d\n", js.number, js.value);
			break;
		}
		pthread_mutex_lock(&joymutex);
		/* Test for motion button */
		if (button[0]) {
			/* compute speed refs */
			printf("\t\tmove %d %d\n", v, w);
			v = axis[1]*VGAIN;
			w = axis[0]*WGAIN;
		} else {
			/* Set refs to zero */
			printf("stop\n");
			v = 0;
			w = 0;
		}
		pthread_mutex_unlock(&joymutex);
	}
}
/*----------------------------------------------------------------------*/
/**
 ** Main loop of joystick control mode
 **/
static void
joymove(struct RMP_DEV_STR **rmp)
{
	unsigned char buf[18];
	int i;
	STATUS retval;

	while (!quit) {
		pthread_mutex_lock(&joymutex);
		for (i = 0; i < 2; i++) {
			if (rmp[i] != NULL)
				retval = rmpSendSpeed(rmp[i], v, w);
			if (retval != OK)
				break;
		}
		pthread_mutex_unlock(&joymutex);
		if (retval != OK) {
			fprintf(stderr, "Are the motors on?\n");
			fprintf(stderr, "aborting.\n");
			quit = 1;
		}
		/* Test for quit button */
		if (button[3])
			quit = 1;
		taskDelay(10);
	} /* while */
} /* joymove */

/*----------------------------------------------------------------------*/

/**
 ** Task reading data from one RMP device.
 ** Frames are decoded and the device state is updated, but not used.
 **/
int
rmpReadTask(int ptr)
{
	struct RMP_DEV_STR *dev = (struct RMP_DEV_STR *)ptr;

	while (!quit) {
		if (rmpReadPackets(dev) < 0)
			break;
		taskDelay(10);
	}
	return OK;
}

/*----------------------------------------------------------------------*/
int
main(int argc, char *argv[])
{
	char name[20];
	struct RMP_DEV_STR **rmp;
	pthread_t t[2];
	int i, j;

	v = 0;
	w = 0;
	quit = 0;

	/* Get a 10ms clock tick */
	h2initGlob(100);

	/* Connect to RMP devices */
	if (rmpOpenAll(&rmp) < 0)
		exit(2);
	printf("rmpOpen done\n");

	/* spawn RMP reader tasks */
	for (i = 0; i < 2; i++) {
		snprintf(name, sizeof(name), "tReadTask%d", i);
		taskSpawn(name, 100, 0, 8192, rmpReadTask, rmp[i]);
	}

	/* Connect to joystick */
	j = joyinit("/dev/input/js0");
	if (j < 0) {
		fprintf(stderr, "Error initializing joystick\n");
		exit(2);
	}
	printf("joyinit done\n");
	axis = calloc(axes, sizeof(int));
	button = calloc(buttons, sizeof(int));

	printf("use button 4 to quit\n"
	       "   button 1 + left stick to move\n");

	taskSpawn("tJoystick", 100, 0, 8192, joyRead, j);
	taskDelay(sysClkRateGet());
	printf("go\n");

	/* launch main loop */
	joymove(rmp);

	/* Free resources */
	free(button);
	free(axis);
	close (j);
	for (i = 0; i < 2; i++) {
		if (rmp[i] != NULL)
			rmpClose(rmp[i]);
	}
	return 0;
}
